/***************************************************************************
    file        : callbacks.c
    project     : gHTTrack (frontend for HTTrack)
    begin       : 2003.05.22
    revised     : 2006.05.08, 2007.01.13
    copyright   : (C)2003-2007 Paul Schuurmans
    email       : paul.schuurmans@home.nl
 ***************************************************************************/

/***************************************************************************
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "jpslib2.h"
#include "main.h"

extern GtkWidget *App1;
extern GtkWidget *FileDlg;
extern TGlobalStruct *Gbl;
extern char *Options[OPTIONCOUNT][4];
extern int CheckDoc[3], FollowRobots[4], Preset[20], RemHost[4], ScanMode[5];
extern char LogType[3], TravelMode[4], TravelModeG[4];

char *CustomDefs[15] = { "", "\%r", "\%h", "\%p/", "\%n", "\%sn", "\%N", "\%sN", "\%[param]", ".\%t", "", "\%M", "\%Q", "\%q", "" };

void on_App1_show (GtkWidget *widget, gpointer user_data)
{
	GtkWidget *w;
	char destpath[260];
	AppDoWidgetSetup();
	// Set destination path (must be done before AppCheckConfigFile)
	sprintf(destpath, "%s/tmp", getenv("HOME"));
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_destination_path");
	gtk_entry_set_text(GTK_ENTRY(w), destpath);
	AppCheckConfigFile();
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_project");
	gtk_combo_box_set_active(GTK_COMBO_BOX(w), Gbl->CurProject);
	strcpy(Gbl->StConfirmed, "No");
	strcpy(Gbl->StCmdLine, "None");
	AppEnableCommands(31);
}

void on_App1_destroy (GtkObject *object, gpointer user_data)
{
	gtk_main_quit();
}

void on_app1_btn_exit_clicked (GtkButton *button, gpointer user_data)
{
	gtk_main_quit();
}

void on_app1_btn_killall_clicked (GtkButton *button, gpointer user_data)
{
	system("killall httrack");
}

void on_app1_btn_perform_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char *astrs[5] = { "Test Links In Pages?", "Download Websites?", "Download All Sites In Pages?", 
		"Continue Interrupted Download?", "Get Specified Files Only?" };
	int ndx = AppGetActionIndex();
	w = lookup_widget(GTK_WIDGET(App1), "confirmdlg_label3");
	gtk_label_set_text(GTK_LABEL(w), astrs[ndx]);
	w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), FALSE);
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_actions");
	gtk_widget_hide(GTK_WIDGET(w));
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_confirm");
	gtk_widget_show(GTK_WIDGET(w));
}

void on_app1_entry_httrack_cmdline_changed (GtkEditable *editable, gpointer user_data)
{
	AppEnableCommands(EC_ACTIONS_PAGE);
}

void on_confirmdlg_btn_cancel_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_confirm");
	gtk_widget_hide(GTK_WIDGET(w));
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_actions");
	gtk_widget_show(GTK_WIDGET(w));
	w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), TRUE);
}

void on_confirmdlg_btn_ok_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char *cmdstrs[5] = { "-p0", "-w", "-Y", "-i", "-g" };
	int ndx;
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_confirm");
	gtk_widget_hide(GTK_WIDGET(w));
	w = lookup_widget(GTK_WIDGET(App1), "app1_frame_actions");
	gtk_widget_show(GTK_WIDGET(w));
	w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), TRUE);
	ndx = AppGetActionIndex();
	strcat(Gbl->CmdLine, cmdstrs[ndx]);
	AppRunCommand();
}

void on_filedlg_btn_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_hide(FileDlg);
}

void on_filedlg_btn_ok_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w = NULL;
	gchar *filename;
	w = lookup_widget(GTK_WIDGET(FileDlg), "filedlg_filechooser");
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(w));
	if(filename == NULL) { g_print("Error: on_filedlg_btn_ok_clicked(), filename=%s\n", filename); return; }
	strcpy(Gbl->FileName, filename);
	g_free(filename);
	gtk_widget_hide(FileDlg);
	if((Gbl->FileDlgMode == FD_URL_LIST_WRITE || Gbl->FileDlgMode == FD_FILTERS_WRITE) && FileExists(Gbl->FileName)) {
		AppSetStatusText("Error: File Exists!  Choose another filename.");
		g_print("File Exists: %s\nChoose another filename.\n", Gbl->FileName);
		return;
	}
	switch(Gbl->FileDlgMode) {
	  case FD_GET_DEST_LOGS:
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_destpath_logs");
		gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
		break;
	  case FD_GET_DEST_PATH:
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_destination_path");
		gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
		break;
	  case FD_FILTERS_READ:
		if(AppValidList(LIST_FILTERS, Gbl->FileName)) {
			w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_filters_file");
			gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
			w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_filters_file");
			ReadWriteWidgetText(FM_READ, w, Gbl->FileName);
		}
		else {
			AppSetStatusText("Error: Invalid Filters List - %s", Gbl->FileName);
			g_print("Error: Invalid Filters List - %s\n", Gbl->FileName);
		}
		break;
	  case FD_FILTERS_WRITE:
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_filters_list");
		ReadWriteWidgetText(FM_WRITE, w, Gbl->FileName);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_filters_file");
		gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_filters_file");
		ReadWriteWidgetText(FM_READ, w, Gbl->FileName);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_filters_file");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
		break;
	  case FD_URL_LIST_READ:
		if(AppValidList(LIST_URLS, Gbl->FileName)) {
			w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_url_listfile");
			gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
			w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_tmp_urls");
			ReadWriteWidgetText(FM_READ, w, Gbl->FileName);
		}
		else {
			AppSetStatusText("Error: Invalid URL List - %s", Gbl->FileName);
			g_print("Error: Invalid URL List - %s\n", Gbl->FileName);
		}
		break;
	  case FD_URL_LIST_WRITE:
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_urls"); 
		ReadWriteWidgetText(FM_WRITE, w, Gbl->FileName);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_url_listfile");
		gtk_entry_set_text(GTK_ENTRY(w), Gbl->FileName);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_rb_use_listfile");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
		break;
	  default: g_print("Error: on_filedlg_btn_ok_clicked() - Unknown mode (%d)\n", Gbl->FileDlgMode); break;
	}
}

void on_logdlg_btn_httrack_help_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char cmdline[280], httrackhelp[260];
	sprintf(httrackhelp, "%s/httrack-cmdline-help.txt", getenv("HOME"));
	if(!FileExists(httrackhelp)) {
		sprintf(cmdline, "httrack --help >> %s", httrackhelp);
		system(cmdline);
	}
	w = lookup_widget(GTK_WIDGET(App1), "logdlg_text_log");
	ReadWriteWidgetText(FM_READ, w, httrackhelp);
}

void on_logdlg_btn_readlog_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char fname[512], path[512];
	gboolean bval = FALSE;
	int nval = 0, len;
	AppGetOptValue("opdlg_entry_destination_path", path, &nval, &bval);
	len = strlen(path);
	if(len) {
		if(path[len-1] != '/') { path[len] = '/'; path[len+1] = 0; }
		sprintf(fname, "%shts-log.txt", path);
		w = lookup_widget(GTK_WIDGET(App1), "logdlg_text_log");
		ReadWriteWidgetText(FM_READ, w, fname);
	}
}

void on_logdlg_btn_wrap_text_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	static int mode = 0;
	int wrapmodes[3] = { GTK_WRAP_NONE, GTK_WRAP_CHAR, GTK_WRAP_WORD };
	if(++mode > GTK_WRAP_WORD)  mode = 0;
	w = lookup_widget(GTK_WIDGET(App1), "logdlg_text_log");
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(w), wrapmodes[mode]);
}

void on_main_notebook_switch_page (GtkNotebook *notebook, GtkNotebookPage *page, guint page_num, gpointer user_data)
{
	if(page_num == 1) {
		strcpy(Gbl->StConfirmed, "No");
		AppEnableCommands(0);
	}
}

void on_opdlg_btn_destination_browse_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_GET_DEST_PATH, "Select Destination Directory...");
}

void on_opdlg_btn_destpath_logs_browse_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_GET_DEST_LOGS, "Select Directory For Logs And Cache...");
}

void on_opdlg_btn_filters_file_browse_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_FILTERS_READ, "Select Filters File...");
}

void on_opdlg_btn_ok_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w, *b;
	char *cbstrs[10] = { "opdlg_cb_use_proxy", "opdlg_cb_run_as_user", "opdlg_cb_use_filters_file", 
		"opdlg_cb_assume_mime", "opdlg_cb_default_referer", "opdlg_cb_from_email_address", 
		"opdlg_cb_use_filters_list", "opdlg_cb_use_filters_file", "opdlg_cb_run_sys_cmd",
		"opdlg_cb_destpath_logs" };
	char *wstrs[10] = { "opdlg_entry_proxy", "opdlg_entry_run_as_user", "opdlg_entry_filters_file", 
		"opdlg_entry_mime1a", "opdlg_entry_default_referer", "opdlg_entry_from_email_address", 
		"opdlg_text_filters_list", "opdlg_entry_filters_file", "opdlg_entry_sys_cmd",
		"opdlg_entry_destpath_logs" };
	int i, nval = 0, err = 0;
	char sval[8192], *txt;
	gboolean bval = FALSE;
	// Unset checkbox if corresponding entry field is empty
	for(i = 0; i < 10; i++) {
		AppGetOptValue(cbstrs[i], sval, &nval, &bval);
		if(bval) {
			AppGetOptValue(wstrs[i], sval, &nval, &bval);
			if(!strlen(sval)) {
				b = lookup_widget(GTK_WIDGET(App1), cbstrs[i]);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), FALSE);
			}
		}
	}
	// Check if opdlg_entry_custom_local_structure not empty
	AppGetOptValue("opdlg_rb_custom_local_structure", sval, &nval, &bval);
	if(bval) {
		AppGetOptValue("opdlg_entry_custom_local_structure", sval, &nval, &bval);
		if(!strlen(sval)) {
			// Non-fatal error: no custom def, so use preset
			b = lookup_widget(GTK_WIDGET(App1), "opdlg_rb_preset_local_structure");
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), TRUE);
		}
	}
	// Check for mandatory URL(s)
	AppGetOptValue("opdlg_rb_use_listfile", sval, &nval, &bval);
	w = lookup_widget(GTK_WIDGET(App1), bval ? "opdlg_text_tmp_urls" : "opdlg_text_urls"); 
	txt = TextViewGetText(w);
	if(strlen(txt)) {
		if(bval) {
			w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_urlfile_editable"); 
			if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w))) {
				w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_tmp_urls"); 
				AppGetOptValue("opdlg_entry_url_listfile", sval, &nval, &bval);
				ReadWriteWidgetText(FM_WRITE, w, sval);
			}
		}
	}
	else  err++;
	if(txt)  g_free(txt);
	// Check for mandatory destination path
	AppGetOptValue("opdlg_entry_destination_path", sval, &nval, &bval);
	if(!strlen(sval))  err++;
	if(err) {
		// Clear all info on Actions page
		w = lookup_widget(GTK_WIDGET(App1), "app1_text_project_info");
		TextViewSetText(w, "");
		strcpy(Gbl->CmdLine, "");
		w = lookup_widget(GTK_WIDGET(App1), "app1_entry_httrack_cmdline");
		gtk_entry_set_text(GTK_ENTRY(w), "");
		AppEnableCommands(EC_ACTIONS_PAGE | EC_OPTIONS_RESET);
		g_print("Error: Missing URL(s) and/or Destination Path.\n");
		AppSetStatusText("Error: Missing URL(s) and/or Destination Path.");
	}
	else {
		// Write current project's settings to config file
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_project");
		AppReadWriteConfig(1, gtk_combo_box_get_active(GTK_COMBO_BOX(w)));
		// Get and set HTTrack command line
		AppShowSettings();
		w = lookup_widget(GTK_WIDGET(App1), "app1_entry_httrack_cmdline");
		gtk_entry_set_text(GTK_ENTRY(w), Gbl->CmdLine);
		// Switch notebook to Actions page
		w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
		gtk_notebook_set_page(GTK_NOTEBOOK(w), 2);
		strcpy(Gbl->StConfirmed, "Yes");
		AppEnableCommands(EC_ACTIONS_PAGE | EC_OPTIONS_RESET);
	}
}

void on_opdlg_btn_quickstart_binary_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char *cbstrs[3] = { "opdlg_cb_get_nonhtml_related", "opdlg_cb_assume_standard", "opdlg_cb_no_purge" };
	char *combostrs[3] = { "opdlg_combo_check_doc_type", "opdlg_combo_follow_robots_txt", "opdlg_combo_travel_mode" };
	int i, comboitems[3] = { 2, 2, 2 };
	for(i = 0; i < 3; i++) {
		w = lookup_widget(GTK_WIDGET(App1), cbstrs[i]);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
		w = lookup_widget(GTK_WIDGET(App1), combostrs[i]);
		gtk_combo_box_set_active(GTK_COMBO_BOX(w), comboitems[i]);
	}
}

void on_opdlg_btn_quickstart_reset_clicked  (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	int i;
	for(i = AppGetOptIndex("opdlg_cb_use_proxy"); i < OPTIONCOUNT; i++) {
		w = lookup_widget(GTK_WIDGET(App1), Options[i][0]);
		switch(Options[i][2][0]) {
		  case 'B':  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);  break;
		  case 'D':  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);  break;
		  case 'E': case 'N':  gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), "");  break;
		  default: break;
		}
	}
}

void on_opdlg_btn_quickstart_spider_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char *cbstrs[3] = { "opdlg_cb_test_validity_all", "opdlg_cb_no_create_index", "app1_rb_action1" };
	int i;
	for(i = 0; i < 3; i++) {
		w = lookup_widget(GTK_WIDGET(App1), cbstrs[i]);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	}
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_cache_updates");
	gtk_combo_box_set_active(GTK_COMBO_BOX(w), 2);
}

void on_opdlg_btn_quickstart_testlinks_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	char *cbstrs[3] = { "opdlg_cb_test_validity_all", "opdlg_cb_no_create_index", "app1_rb_action1" };
	int i;
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_mirror_depth");
	gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), "1");
	for(i = 0; i < 3; i++) {
		w = lookup_widget(GTK_WIDGET(App1), cbstrs[i]);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
	}
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_cache_updates");
	gtk_combo_box_set_active(GTK_COMBO_BOX(w), 2);
}

void on_opdlg_btn_reset_defaults_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *w;
	int i, nval = 0, pos = 0, prj = Gbl->CurProject;
	char *sval, configfile[260], key[80];
	gboolean bval = FALSE;
	FILE *fn = NULL;
	if(prj) {
		sprintf(configfile, "%s/.ghttrack.cfg", getenv("HOME"));
		if((fn = fopen(configfile, "rb")) == NULL) {
			g_print("reset_defaults(), fopen(%s, \"rb\") failed, fn=%p\n", configfile, fn);
			return;
		}
	}
	sval = g_malloc0(4096);
	for(i = 0; i < OPTIONCOUNT; i++) {
		w = lookup_widget(GTK_WIDGET(App1), Options[i][0]);
		// If project is "Custom Settings n", set to settings in "Default Settings".
		// Else, reset to HTTrack defaults (i.e., clear the current project's settings).
		if(prj) {
			sprintf(key, "0%s", Options[i][0]); 
			switch(Options[i][2][0]) {
			  case 'B':
				GetConfigBool(fn, key, &bval, FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), bval);
				break;
			  case 'D':
				GetConfigInt(fn, key, &nval, 0);
				gtk_combo_box_set_active(GTK_COMBO_BOX(w), nval);
				break;
			  case 'E':
				GetConfigString(fn, key, sval, "");
				gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), sval);
				break;
			  case 'N':
				GetConfigInt(fn, key, &nval, 0);
				if(nval == -1)  strcpy(sval, "");
				else  sprintf(sval, "%d", nval);
				gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), sval);
				break;
			  case 'S':
				GetConfigString(fn, key, sval, "");
				gtk_entry_set_text(GTK_ENTRY(w), sval);
				break;
			  case 'T':
				GetConfigString(fn, key, sval, "");
				for(pos = 0; pos < strlen(sval); pos++)  if(sval[pos] == ' ')  sval[pos] = 10;
				TextViewSetText(w, sval);
				break;
			  default: break;
			}
		}
		else {
			switch(Options[i][2][0]) {
			  case 'B':  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);  break;
			  case 'D':  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);  break;
			  case 'E': case 'N':  gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), "");  break;
			  case 'S':  gtk_entry_set_text(GTK_ENTRY(w), "");  break;
			  case 'T':  TextViewSetText(w, "");  break;
			  default: break;
			}
		}
	}
	if(fn)  fclose(fn);
	g_free(sval);
}

void on_opdlg_btn_save_filters_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_FILTERS_WRITE, "Save Filters List To File...");
}

void on_opdlg_btn_save_listfile_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_URL_LIST_WRITE, "Save URL List To File...");
}

void on_opdlg_btn_save_project_clicked  (GtkButton *button, gpointer user_data)
{
	AppReadWriteConfig(FM_WRITE, Gbl->CurProject);
}

void on_opdlg_btn_url_listfile_browse_clicked (GtkButton *button, gpointer user_data)
{
	AppRunFileDialog(FD_URL_LIST_READ, "Select URL List File...");
}

void on_opdlg_cb_assume_mime_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE3);
}

void on_opdlg_cb_create_word_db_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_no_create_index");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
}

void on_opdlg_cb_debug_mode_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_create_logs");
		if(gtk_combo_box_get_active(GTK_COMBO_BOX(w)) == 3)  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
	}
}

void on_opdlg_cb_default_referer_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE3);
}

void on_opdlg_cb_destpath_logs_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_cb_from_email_address_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE3);
}

void on_opdlg_cb_no_create_index_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_create_word_db");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_top_index");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
}

void on_opdlg_cb_run_as_user_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_cb_run_sys_cmd_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE5);
}

void on_opdlg_cb_store_all_in_cache_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_cache_updates");
		if(gtk_combo_box_get_active(GTK_COMBO_BOX(w)) == 2)  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
	}
}

void on_opdlg_cb_top_index_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_no_create_index");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
}

void on_opdlg_cb_urlfile_editable_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	gboolean editable, bval;
	char filename[260] = "";
	int nval;
	editable = gtk_toggle_button_get_active(togglebutton);
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_tmp_urls");
	gtk_text_view_set_editable(GTK_TEXT_VIEW(w), editable);
	if(editable == FALSE) {
		// Restore original file contents
		AppGetOptValue("opdlg_entry_url_listfile", filename, &nval, &bval);
		ReadWriteWidgetText(FM_READ, w, filename);
	}
}

void on_opdlg_cb_use_filters_file_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_filters_list");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
	AppEnableCommands(EC_OPTIONS_PAGE5);
}

void on_opdlg_cb_use_filters_list_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_toggle_button_get_active(togglebutton)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_filters_file");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
	AppEnableCommands(EC_OPTIONS_PAGE5);
}

void on_opdlg_cb_use_hostname_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_cb_use_proxy_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_combo_cache_updates_changed (GtkComboBox *combobox, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_combo_box_get_active(combobox) == 2) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_store_all_in_cache");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
}

void on_opdlg_combo_create_logs_changed (GtkComboBox *combobox, gpointer user_data)
{
	GtkWidget *w;
	if(gtk_combo_box_get_active(combobox) == 3) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_debug_mode");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
	}
}

void on_opdlg_combo_custom_struct_defs_changed (GtkComboBox *combobox, gpointer user_data)
{
	GtkWidget *w;
	int ndx = 0;
	static int pos = 0;
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_custom_struct_defs");
	ndx = gtk_combo_box_get_active(GTK_COMBO_BOX(w));
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_custom_local_structure");
	switch(ndx) {
	  case 0: case 10: case 14: case 16:  break;
	  case 15:  gtk_entry_set_text(GTK_ENTRY(w), ""); break;
	  default:
		gtk_editable_insert_text(GTK_EDITABLE(w), CustomDefs[ndx], strlen(CustomDefs[ndx]), &pos);
		break;
	};
}

void on_opdlg_combo_project_changed (GtkComboBox *combobox, gpointer user_data)
{
	GtkWidget *w;
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_project");
	Gbl->CurProject = gtk_combo_box_get_active(GTK_COMBO_BOX(w));
	AppReadWriteConfig(FM_READ, Gbl->CurProject);
	AppEnableCommands(EC_OPTIONS_RESET);
}

void on_opdlg_entry_destination_path_changed (GtkEditable *editable, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_entry_url_listfile_changed (GtkEditable *editable, gpointer user_data)
{
	GtkWidget *w;
	char filename[260] = "";
	strcpy(filename, gtk_entry_get_text(GTK_ENTRY(editable)));
	if(strlen(filename)) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_tmp_urls");
		ReadWriteWidgetText(FM_READ, w, filename);
	}
	AppEnableCommands(EC_OPTIONS_PAGE1);
}

void on_opdlg_rb_custom_local_structure_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE4);
}

void on_opdlg_rb_preset_local_structure_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE4);
}

void on_opdlg_rb_use_list_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	AppEnableCommands(EC_OPTIONS_PAGE1);
}


