/***************************************************************************
    file        : jpslib2.c
    project     : JPS Library for GTK2
    begin       : 2007.01.11
    revised     : --
    copyright   : (C)2007 Paul Schuurmans
    email       : paul.schuurmans@home.nl
 ***************************************************************************/

/***************************************************************************
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ***************************************************************************/

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ctype.h>
#include <fcntl.h>
#include <utime.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <utime.h>

#include "support.h"
#include "jpslib2.h"

int FileExists(char *path)
{
	return(access(path, 0) == 0);
}

long FileSize (char *name)
{
	long eof_ftell;
	FILE *file;
	file = fopen(name, "r");
	if(file == NULL)  return(0L);
	fseek(file, 0L, SEEK_END);
	eof_ftell = ftell(file);
	fclose(file);
	return(eof_ftell);
}

void GetConfigBool (FILE *fn, char *key, gboolean *bval, gboolean bdef)
{
	char line[1024] = "", txt[1024] = "";
	int len = strlen(key);
	if(fn != NULL) {
		fseek(fn, 0, SEEK_SET);
		while(ReadLine(fn, line, 1023) != -1) {
			if(!strncmp(key, line, len)) {
				strcpy(txt, line+len+1);
				if(!strcasecmp(txt, "true") || !strcmp(txt, "1")) *bval = TRUE;
				else  *bval = FALSE;
				break;
			}
		}
	}
	else  *bval = bdef;
}

void GetConfigInt (FILE *fn, char *key, int *nval, int ndef)
{
	char line[1024] = "", txt[1024] = "";
	int len, i, keylen = strlen(key);
	if(fn != NULL) {
		fseek(fn, 0, SEEK_SET);
		while(ReadLine(fn, line, 1023) != -1) {
			if(!strncmp(key, line, keylen)) {
				strcpy(txt, line+keylen+1);
				len = strlen(txt);
				for(i = 0; i < len; i++) {
					switch(txt[i]) {
					  case '0': case '1': case '2': case '3': case '4': case '5': 
					  case '6': case '7': case '8': case '9':  break;
					  default:  len = 0;  break;
					}
				}
				if(len)  *nval = atoi(txt);
				else  *nval = -1;
				break;
			}
		}
	}
	else  *nval = ndef;
}

void GetConfigString (FILE *fn, char *key, char *sval, char *sdef)
{
	char line[1024] = "";
	int len = strlen(key);
	if(fn != NULL) {
		fseek(fn, 0, SEEK_SET);
		while(ReadLine(fn, line, 1023) != -1) {
			if(!strncmp(key, line, len)) {
				strcpy(sval, line+len+1);
				break;
			}
		}
	}
	else {
		if(strlen(sdef))  strcpy(sval, sdef);
		else  strcpy(sval, "");
	}
}

int ReadLine (FILE *rfile, char *line, int maxlinesize)
{
	int ch, c = 0;
	do {
		ch = getc(rfile);
		if((ch == EOF) || (ch == -1)) return(-1);
		if(ch == 13)  continue;
		if(ch == 10) break;
		line[c] = ch;
		c++;
	} while(c < maxlinesize && ch != 10);
	line[c] = 0;
	return(c);
}

int ReadWriteWidgetText (int mode, GtkWidget *w, char *filename)
{
	FILE *file;
	gchar *txt = NULL;
	char *modestr[2] = { "rb", "wb" };
	int fsize = 0;
	if((file = fopen(filename, modestr[mode])) != NULL) {
		if(mode == FM_WRITE) {
			if((txt = TextViewGetText(w)) != NULL)  fwrite(txt, strlen(txt), 1, file);
			else { g_print("Error: ReadWriteWidgetText(%d, %p, %s)\n", mode, w, filename); return(0); }
		}
		else {
			fsize = FileSize(filename);
			txt = g_malloc(fsize+1);
			fread(txt, fsize, 1, file);
			txt[fsize] = 0;
			TextViewSetText(w, txt);
		}
		fclose(file);
		g_free(txt);
		return(1);
	}
	return(0);
}

gchar *TextViewGetText (GtkWidget *w)
{
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(w));
	gtk_text_buffer_get_start_iter(buffer, &start);
	gtk_text_buffer_get_end_iter(buffer, &end);
	return(gtk_text_buffer_get_text(buffer, &start, &end, FALSE));
}

void TextViewSetText (GtkWidget *w, gchar *text)
{
	GtkTextBuffer *buffer;
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(w));
	gtk_text_buffer_set_text(buffer, text, -1);
}

int WidgetIsEnabled (GtkWidget *parent, char *name)
{
	GtkWidget *w = NULL;
	w = lookup_widget(GTK_WIDGET(parent), name);
	if(!w)  return(-1);
	if(GTK_WIDGET_IS_SENSITIVE(w))  return(1);
	return(0);
}
