/***************************************************************************
    file        : main.c
    project     : gHTTrack (frontend for HTTrack)
    begin       : 2003.05.22
    revised     : 2006.05.08
                  2007.01.13 (0.5.2)
    copyright   : (C)2003-2007 Paul Schuurmans
    email       : paul.schuurmans@home.nl
 ***************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ctype.h>
#include <fcntl.h>
#include <utime.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <utime.h>

#include "interface.h"
#include "support.h"
#include "callbacks.h"
#include "jpslib2.h"
#include "main.h"

GtkWidget *App1;
GtkWidget *FileDlg;
TGlobalStruct *Gbl;
pid_t ChildPid;

char *Options[OPTIONCOUNT][4] = {
	// Line format: Name, Switch, Type, Description
	// Codes for Type: S=string, T=text, N=number, B=boolean, D=digit
	// 0-3, 1st tabpage, URLs
	{ "opdlg_cb_run_quiet", "q", "B", "Run HTTrack In Quiet Mode" },
	{ "opdlg_text_urls", "", "T2047", "URLs" },
	{ "opdlg_rb_use_listfile", "", "B", "Use URL List File" },
	{ "opdlg_entry_url_listfile", "\%L $", "S255", "URL List File" },
	// 4-6, 1st tabpage, Destination Path
	{ "opdlg_entry_destination_path", "O $", "S255", "Destination Path" },
	{ "opdlg_cb_destpath_logs", "", "B", "Store Logs And Cache Separately" },
	{ "opdlg_entry_destpath_logs", "", "S255", "Destination Path For Logs And Cache" },
	// 7-10, 1st tabpage, Proxy Server
	{ "opdlg_cb_use_proxy", "", "B", "Use Proxy" },
	{ "opdlg_entry_proxy", "P $", "S255", "Proxy" },
	{ "opdlg_entry_proxy_port", ":$", "S7", "Port" },
	{ "opdlg_cb_no_ftp_proxy", "\%f0", "B", "No Proxy For FTP" },
	// 11-14, 1st tabpage, Miscellaneous
	{ "opdlg_cb_run_as_user", "", "B", "Run As User" },
	{ "opdlg_entry_run_as_user", "\%U $", "S31", "User" },
	{ "opdlg_cb_use_hostname", "", "B", "Run As Host" },
	{ "opdlg_entry_hostname", "\%b $", "S80", "Hostname" },
	// 15-24, 2nd tabpage, Limits
	{ "opdlg_entry_mirror_depth", "r#", "N6", "Max. Mirror Depth" },
	{ "opdlg_entry_external_depth", "\%e#", "N4", "Max. External Depth" },
	{ "opdlg_entry_max_size_html", ",m#", "N8", "Max. Size HTML" },
	{ "opdlg_entry_max_size_nonhtml", "m#", "N10", "Max. Size Non-HTML" },
	{ "opdlg_entry_site_size_limit", "M#", "N10", "Site Size Limit" },
	{ "opdlg_entry_pause_after", "G#", "N10", "Pause After Downloading" },
	{ "opdlg_entry_max_time_overall", "E#", "N8", "Max. Time Overall" },
	{ "opdlg_entry_transfer_rate", "A#", "N10", "Max. Transfer Rate" },
	{ "opdlg_entry_max_connections", "\%c#", "N4", "Max. Connections" },
	{ "opdlg_entry_max_links", "#L#", "N10", "Max. Links" },	// 1st '#' is part of command option
	// 25-29, 2nd tabpage, Flow Control
	{ "opdlg_entry_connections", "c#", "N4", "Number of Connections" },
	{ "opdlg_entry_timeout", "T#", "N6", "Timeout" },
	{ "opdlg_entry_retries", "R#", "N4", "Number Of Retries" },
	{ "opdlg_entry_min_transfer_rate", "J#", "N8", "Minimum Transfer Rate" },
	{ "opdlg_combo_remove_host", "H#", "D", "Abandon Host" },	// H1=Timeout,H2=TrRateLow,H3=Both
	// 30-32, 2nd tabpage, Links
	{ "opdlg_cb_no_extended_parsing", "\%P0", "B", "No Extended Parsing" },
	{ "opdlg_cb_get_nonhtml_related", "n", "B", "Get Non-HTML Related To Link" },
	{ "opdlg_cb_test_validity_all", "t", "B", "Test All Links" },
	// 33-43, 3rd tabpage, Spider
	{ "opdlg_cb_no_cookies", "b0", "B", "Do Not Accept Cookies" },
	{ "opdlg_cb_no_java", "j0", "B", "Do Not Parse Java Classes" },
	{ "opdlg_cb_force_http1", "\%h", "B", "Force HTTP/1.0" },
	{ "opdlg_cb_tolerant_requests", "\%B", "B", "Tolerant Requests" },
	{ "opdlg_cb_use_update_hacks", "\%s", "B", "Use Update Hacks" },
	{ "opdlg_cb_use_url_hacks", "\%u", "B", "Use URL Hacks" },
	{ "opdlg_cb_keep_alive", "\%k", "B", "Use Keep-Alive" },
	{ "opdlg_cb_assume_standard", "-assume standard", "B", "Assume Standard MIME Types" },
	{ "opdlg_combo_check_doc_type", "u#", "D", "Check Document Type" },
	{ "opdlg_combo_follow_robots_txt", "s#", "D", "Follow robots.txt" },
	{ "opdlg_combo_internet_protocol", "@i#", "D", "Internet Protocol" },
	// 44-60, 3rd tabpage, MIME Associations
	{ "opdlg_cb_assume_mime", "\%A $", "B", "Assume MIME Type Links" },
	{ "opdlg_entry_mime1a", "", "S31", "MimeAssociation1a" },
	{ "opdlg_entry_mime1b", "", "S31", "MimeAssociation1b" },
	{ "opdlg_entry_mime2a", "", "S31", "MimeAssociation2a" },
	{ "opdlg_entry_mime2b", "", "S31", "MimeAssociation2b" },
	{ "opdlg_entry_mime3a", "", "S31", "MimeAssociation3a" },
	{ "opdlg_entry_mime3b", "", "S31", "MimeAssociation3b" },
	{ "opdlg_entry_mime4a", "", "S31", "MimeAssociation4a" },
	{ "opdlg_entry_mime4b", "", "S31", "MimeAssociation4b" },
	{ "opdlg_entry_mime5a", "", "S31", "MimeAssociation5a" },
	{ "opdlg_entry_mime5b", "", "S31", "MimeAssociation5b" },
	{ "opdlg_entry_mime6a", "", "S31", "MimeAssociation6a" },
	{ "opdlg_entry_mime6b", "", "S31", "MimeAssociation6b" },
	{ "opdlg_entry_mime7a", "", "S31", "MimeAssociation7a" },
	{ "opdlg_entry_mime7b", "", "S31", "MimeAssociation7b" },
	{ "opdlg_entry_mime8a", "", "S31", "MimeAssociation8a" },
	{ "opdlg_entry_mime8b", "", "S31", "MimeAssociation8b" },
	// 61-66, 3rd tabpage, Browser ID
	{ "opdlg_entry_browser_identity", "F $", "E255", "Browser Identity" },
	{ "opdlg_entry_preferred_language", "\%l $", "E7", "Preferred Language" },
	{ "opdlg_cb_default_referer", "", "B", "Use Default Referer" },
	{ "opdlg_entry_default_referer", "\%R $", "S255", "Default Referer" },
	{ "opdlg_cb_from_email_address", "", "B", "Use From Email Address" },
	{ "opdlg_entry_from_email_address", "\%E $", "S79", "From Email Address" },
	// 67-78, 4th tabpage, Local Structure
	{ "opdlg_rb_preset_local_structure", "", "B", "Use Preset Local Structure" },
	{ "opdlg_combo_preset_lstruct", "N#", "D", "Local Structure" },	// choices: 0-5,99-105,199,1001-1005,1099
	{ "opdlg_rb_custom_local_structure", "", "B", "Use User-Defined Local Structure" },
	{ "opdlg_entry_custom_local_structure", "N $", "S255", "User-Defined Local Structure" },
	{ "opdlg_cb_no_external_pages", "x", "B", "No External Pages" },
	{ "opdlg_cb_hide_passwords", "\%x", "B", "Hide Passwords" },
	{ "opdlg_cb_hide_query_strings", "\%q0", "B", "Hide Query Strings" },
	{ "opdlg_cb_suppress_errors", "o", "B", "Suppress HTML Error Files" },
	{ "opdlg_cb_no_purge", "X0", "B", "Do Not Purge Old Files" },
	{ "opdlg_cb_preserve_as_is", "\%p", "B", "Preserve HTML Files 'As Is'" },
	{ "opdlg_combo_long_names", "L#", "D", "Long Names" },
	{ "opdlg_combo_keep_original", "", "D", "Keep Original Links" },
	// 79, 4th tabpage, HTML Footer
	{ "opdlg_entry_html_footer", "\%F $", "E255", "HTML Footer" },
	// 80-87, 4th tabpage, Log/Index/Cache
	{ "opdlg_combo_create_logs", "", "D", "Log Type" },		// choices: z,Z,Q
	{ "opdlg_combo_cache_updates", "C#", "D", "Cache Updates" },		// choices: 2,1,0
	{ "opdlg_cb_debug_mode", "\%H", "B", "Activate Debug Mode" },
	{ "opdlg_cb_no_create_index", "I0", "B", "Do Not Create Index (index.html)" },
	{ "opdlg_cb_create_word_db", "\%I", "B", "Create Searchable Index File" },
	{ "opdlg_cb_store_all_in_cache", "k", "B", "Store All Files in Cache" },
	{ "opdlg_cb_no_reget_deleted", "\%n", "B", "Do Not Reget Locally Deleted Files" },
	{ "opdlg_cb_top_index", "\%i", "B", "Make A Top Index For A Project Folder" },
	// 88-91, 4th tabpage, Scan Rules
	{ "opdlg_cb_use_filters_list", "", "B", "Use Scan Rules List" },
	{ "opdlg_text_filters_list", "", "T2047", "Scan Rules List" },
	{ "opdlg_cb_use_filters_file", "", "B", "Use Scan Rules File" },
	{ "opdlg_entry_filters_file", "\%S $", "S255", "Scan Rules File" },
	// 92-94, 4th tabpage, Expert
	{ "opdlg_combo_scan_mode", "p#", "D", "Primary Scan Rules" },		// choices: 0,1,2,3*,7
	{ "opdlg_combo_travel_mode", "", "D", "Travel Mode" },				// choices: S,D*,U,B
	{ "opdlg_combo_global_travel_mode", "", "D", "Global Travel Mode" },	// choices: a*,d,l,e
	// 95-97, 4th tabpage, Guru
	{ "opdlg_cb_flush_logs", "#f", "B", "Always Flush Log Files" },
	{ "opdlg_cb_generate_transfer_log", "#T", "B", "Generate Transfer Log" },
	{ "opdlg_entry_max_filters", "#F#", "N4", "Maximum Filters" },
	// 98-99, 4th tabpage, Miscellaneous
	{ "opdlg_cb_run_sys_cmd", "", "B", "Execute System Command After Each Retrieved File" },
	{ "opdlg_entry_sys_cmd", "V $", "S255", "System Command To Execute" }
};

int CacheUd[3] = { 2, 1, 0 };
int CheckDoc[3] = { 1, 2, 0 };
int FollowRobots[4] = { 2, 3, 1, 0 };
int InternetProtocol[3] = { 0, 4, 6 };
char *KeepOriginal[4] = { "-K0", "-K3", "-K", "-K4" };
char LogType[4] = { ' ', 'z', 'Z', 'Q' };
int LongNames[3] = { 1, 0, 2 };
int Preset[20] = { 0, 1, 2, 3, 4, 5, 99, 100, 101, 102, 103, 104, 105, 199, 1001, 1002, 1003, 1004, 1005, 1099 };
int RemHost[4] = { 0, 1, 2, 3 };
int ScanMode[5] = { 3, 0, 1, 2, 7 };
char TravelMode[4] = { 'D', 'U', 'B', 'S' };
char TravelModeG[4] = { 'a', 'd', 'l', 'e' };

///// Application-Specific Functions ////////////////////////////////////////

void AppCheckConfigFile ()
{
	GtkWidget *w;
	gboolean bval = FALSE, writeconfigfile = TRUE;
	FILE *fn;
	int i, n, scnt = 0, nval = 0;
	char configfile[260] = "", *sval, tmpstr[8192];
	sval = g_malloc0(8192);
	sprintf(configfile, "%s/.ghttrack.cfg", getenv("HOME"));
	if(FileExists(configfile)) {
		if((fn = fopen(configfile, "rb")) == NULL) {
			AppSetStatusText("AppCheckConfigFile(), fopen(%s, \"rb\") failed, using defaults, fn=%p", configfile, fn);
			g_free(sval);
			return;
		}
		AppSetStatusText("AppCheckConfigFile(), config found, reading %s...", configfile);
		ReadLine(fn, tmpstr, 8190);
		if(!strcmp(tmpstr, CFGID01)) {
			GetConfigInt(fn, "CurProject", &Gbl->CurProject, 0);
			fclose(fn);
			AppReadWriteConfig(FM_READ, Gbl->CurProject);
			w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
			gtk_notebook_set_page(GTK_NOTEBOOK(w), 1);
			writeconfigfile = FALSE;
		}
	}
	if(writeconfigfile) {
		//configfile not found or is old version
		if((fn = fopen(configfile, "w+")) == NULL) {
			AppSetStatusText("Error: AppCheckConfigFile(), fopen(%s, \"w+\") failed, fn=%p", configfile, fn);
			g_print("Error: AppCheckConfigFile(), fopen(%s, \"w+\") failed, fn=%p\n", configfile, fn);
			g_free(sval);
			return;
		}
		fprintf(fn, "%s\n", CFGID01);
		fprintf(fn, "CurProject=%d\n", Gbl->CurProject);
		for(scnt = 0; scnt < 10; scnt++) {
			for(i = 0; i < OPTIONCOUNT; i++) {
				AppGetOptValue(Options[i][0], sval, &nval, &bval);
				switch(Options[i][2][0]) {
				  case 'S': case 'T': case 'E':
					strcpy(tmpstr, sval);
					for(n = 0; n < strlen(tmpstr); n++)  if(tmpstr[n] == 10 || tmpstr[n] == 13)  tmpstr[n] = ' ';
					fprintf(fn, "%d%s=%s\n", scnt, Options[i][0], tmpstr);
					break;
				  case 'N': case 'D':  fprintf(fn, "%d%s=%d\n", scnt, Options[i][0], nval);  break;
				  case 'B':  fprintf(fn, "%d%s=%d\n", scnt, Options[i][0], bval);  break;
				  default: break;
				}
			}
		}
		fclose(fn);
	}
	g_free(sval);
}

void AppDoWidgetSetup ()
{
	GtkWidget *w;
	GtkTooltipsData *tipdata;
	char numstr[16];
	char *opt[32][2] = {
		{ "opdlg_entry_mirror_depth", "Maximum directory depth (N levels deep)." },
		{ "opdlg_entry_external_depth", "Depth the engine will seek in external sites, or on addresses that were forbidden. Note: Use this option with great care, as it overrides all other options (filters and default engine limiter)." },
		{ "opdlg_entry_max_size_html", "Maximum file length (in bytes) for HTML files." },
		{ "opdlg_entry_max_size_nonhtml", "Maximum file length (in bytes) for non-HTML files." },
		{ "opdlg_entry_site_size_limit", "The total amount of bytes that can be downloaded in the current mirror." },
		{ "opdlg_entry_pause_after", "Pause every time the engine has retrieved N bytes, and wait until lock file is deleted." },
		{ "opdlg_entry_max_time_overall", "The total amount of time (in seconds) that can be spent on the current mirror." },
		{ "opdlg_entry_transfer_rate", "Maximum transfer rate in bytes per second." },
		{ "opdlg_entry_max_connections", "The number of connections per second for the current mirror." },
		{ "opdlg_entry_max_links", "Maximum number of links. Note: This is a guru option and should not be used, if possible." },
		{ "opdlg_entry_connections", "Number of simultaneous downloads to allow." },
		{ "opdlg_entry_timeout", "Number of seconds after a non-responding link is shut down." },
		{ "opdlg_entry_retries", "Number of retries, in case of timeout or non-fatal errors." },
		{ "opdlg_entry_min_transfer_rate", "Traffic jam control. Minimum transfer rate (in bytes per second) tolerated for a link." },
		{ "opdlg_combo_remove_host", "When to abandon the host." },
		{ "opdlg_combo_check_doc_type", "Define when to check document type.  The engine must know the document type, to rewrite the file types. For example, if a link called /cgi-bin/gen_image.cgi generates a gif image, the generated file will not be called \"gen_image.cgi\" but \"gen_image.gif\".  Note: Avoid \"never\", because the local mirror could be bogus." },
		{ "opdlg_combo_follow_robots_txt", "Define whether the engine should follow remote robots.txt rules, if they exist." },
		{ "opdlg_combo_internet_protocol", "Internet protocol(s) to use." },
		{ "opdlg_entry_browser_identity", "How you want HTTrack to identify itself to Web servers.  Some servers use this information to determine how they should display their page content." },
		{ "opdlg_entry_preferred_language", "Specify, to Web servers, the preferred language.  Some servers use this information to determine how they should display their page content." },
		{ "opdlg_combo_preset_lstruct", "How files should be stored on the local mirror." },
		{ "opdlg_combo_custom_struct_defs", "Select an item to insert into the custom definition (or select a blank line to leave unchanged)." },
		{ "opdlg_combo_long_names", "The filename convention to use." },
		{ "opdlg_combo_keep_original", "How the original links in files should be kept on the local mirror." },
		{ "opdlg_entry_html_footer", "The information that will be included at the bottom of each retrieved HTML page." },
		{ "opdlg_combo_create_logs", "The type of logging to use." },
		{ "opdlg_combo_scan_mode", "Primary Scan Rules." },
		{ "opdlg_combo_travel_mode", "Travel Mode." },
		{ "opdlg_combo_global_travel_mode", "Global Travel Mode." },
		{ "opdlg_combo_cache_updates", "Create/use a cache for updates and retries." },
		{ "opdlg_entry_max_filters", "The maximum number of filters to use." },
		{ "opdlg_combo_project", "The current project.  You can set common options in the 'Default Settings' project and use it as a template for other projects.  Note: If you switch to another project, you will lose any changes made to the current project unless you press [Save] before switching." }
	};
	int i;
	// Set limits for those not supported in Glade2
	for(i = 0; i < OPTIONCOUNT; i++) {
		if(Options[i][2][0] == 'N') {
			w = lookup_widget(GTK_WIDGET(App1), Options[i][0]);
			strcpy(numstr, Options[i][2] + 1);
			gtk_entry_set_max_length(GTK_ENTRY(GTK_BIN(w)->child), atoi(numstr));
		}
	}
	// Set tooltips for those not supported in Glade2
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_url_listfile");
	tipdata = gtk_tooltips_data_get(w);
	for(i = 0; i < 32; i++) {
		w = lookup_widget(GTK_WIDGET(App1), opt[i][0]);
		gtk_tooltips_set_tip(tipdata->tooltips, GTK_BIN(w)->child, opt[i][1], "");
		if(opt[i][0][6] == 'c')  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
	}
}

void AppEnableCommands (int mode)
{
	GtkWidget *w, *cmdline, *itm, *t1, *t2;
	char txt[8192] ="";
	gchar *text;
	gboolean enable = FALSE;
	int len = 0;
	if(mode & EC_ACTIONS_PAGE) {
		cmdline = lookup_widget(GTK_WIDGET(App1), "app1_entry_httrack_cmdline");
		strcpy(txt, gtk_entry_get_text(GTK_ENTRY(cmdline)));
		if((strlen(txt) > 6) && (strncmp(txt, "httrack ", 7) == 0))  enable = TRUE;
		strcpy(Gbl->StCmdLine, enable ? "Yes" : "No");
		itm = lookup_widget(GTK_WIDGET(App1), "app1_vbox_action_rbs");
		gtk_widget_set_sensitive(GTK_WIDGET(itm), enable);
		itm = lookup_widget(GTK_WIDGET(App1), "app1_btn_perform");
		gtk_widget_set_sensitive(GTK_WIDGET(itm), enable);
	}
	if(mode & EC_OPTIONS_PAGE1) {
		// URLs list
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_rb_use_list");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		t1 = lookup_widget(GTK_WIDGET(App1), "opdlg_vbox_use_url_list");
		t2 = lookup_widget(GTK_WIDGET(App1), "opdlg_vbox_use_url_file");
		gtk_widget_hide(GTK_WIDGET(enable ? t2 : t1));
		gtk_widget_show(GTK_WIDGET(enable ? t1 : t2));
		w = lookup_widget(GTK_WIDGET(App1), enable ? "opdlg_text_urls" : "opdlg_text_tmp_urls");
		text = TextViewGetText(w);
		len = strlen(text);
		g_free(text);
		strcpy(Gbl->StUrls, len ? "Yes" : "None");
		//proxy server
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_proxy");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_proxy");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable ? TRUE : FALSE);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_proxy_port");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable ? TRUE : FALSE);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_no_ftp_proxy");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable ? TRUE : FALSE);
		//destpath
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_destination_path");
		strcpy(txt, gtk_entry_get_text(GTK_ENTRY(w)));
		strcpy(Gbl->StDestPath, strlen(txt) ? "Yes" : "No");
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_destpath_logs");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_destpath_logs");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_btn_destpath_logs_browse");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		//run as user
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_run_as_user");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_run_as_user");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_hostname");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_hostname");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
	}
	if(mode & EC_OPTIONS_PAGE3) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_assume_mime");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_table_mime_defs");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_default_referer");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_default_referer");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_from_email_address");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_from_email_address");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
	}
	if(mode & EC_OPTIONS_PAGE4) {
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_rb_preset_local_structure");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_preset_lstruct");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_hbox_custom_lstruct");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable ? FALSE : TRUE);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_custom_struct_defs");
		gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
	}
	if(mode & EC_OPTIONS_PAGE5) {
		// Filters list
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_filters_list");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_btn_save_filters");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_filters_list");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		// Filters file
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_use_filters_file");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_btn_filters_file_browse");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_filters_file");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_text_filters_file");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
		// Miscellaneous
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_run_sys_cmd");
		enable = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_sys_cmd");
		gtk_widget_set_sensitive(GTK_WIDGET(w), enable);
	}
	if(mode & EC_OPTIONS_RESET) {
		// Reset opdlg_cb_urlfile_editable checkbox
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_cb_urlfile_editable");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
		// Reset opdlg_combo_custom_struct_defs dropdown
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_custom_struct_defs");
		gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
	}
	if(ChildPid)  AppSetStatusText("URLs : %s    DestinationPath : %s    OptionsConfirmed : %s    CommandLine : %s    ProcessId : %d", 
	  Gbl->StUrls, Gbl->StDestPath, Gbl->StConfirmed, Gbl->StCmdLine, ChildPid);
	else  AppSetStatusText("URLs : %s    DestinationPath : %s    OptionsConfirmed : %s    CommandLine : %s    ProcessId : NA", 
	  Gbl->StUrls, Gbl->StDestPath, Gbl->StConfirmed, Gbl->StCmdLine);
}

int AppGetActionIndex ()
{
	GtkWidget *w;
	char *rbstrs[5] = { "app1_rb_action1", "app1_rb_action2", "app1_rb_action3", "app1_rb_action4", "app1_rb_action5" };
	int i;
	for(i = 0; i < 5; i++) {
		w = lookup_widget(GTK_WIDGET(App1), rbstrs[i]);
		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)))  break;
	}
	return(i);
}

int AppGetOptIndex (char *name)
{
	int i, ndx = -1;
	for(i = 0; i < OPTIONCOUNT; i++) {
		if(!strcmp(name, Options[i][0])) {
			ndx = i;
			break;
		}
	}
	return(ndx);
}

char *AppGetOptSwitch (char *name, char *cmdsw)
{
	int i, end = 0;
	for(i = 0; i < OPTIONCOUNT; i++) {
		if(!strcmp(name, Options[i][0])) {
			sprintf(cmdsw, "-%s", Options[i][1]);
			end = strlen(cmdsw) - 1;
			if(cmdsw[end] == '$' || cmdsw[end] == '#' || cmdsw[end] == '-')  cmdsw[end] = 0;
			return(cmdsw);
		}
	}
	cmdsw[0] = 0;
	return(cmdsw);
}

int AppGetOptValue (char *name, char *sval, int *nval, gboolean *bval)
{
	GtkWidget *w;
	gchar *txt = NULL;
	GtkTextBuffer *textbuffer;
	GtkTextIter start, end;
	int i, ndx = -1, len = 0, retval = 0;
	for(i = 0; i < OPTIONCOUNT; i++) {
		if(!strcmp(name, Options[i][0])) {
			ndx = i;
			break;
		}
	}
	if(ndx == -1 || (w = lookup_widget(GTK_WIDGET(App1), name)) == 0) { 
		g_print("Error: AppGetOptValue(%s, %s, %d, %d)\n", name, sval, *nval, *bval); 
		return(0); 
	}
	switch(Options[ndx][2][0]) {
	  case 'B':
		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)))  *bval = TRUE;
		else  *bval = FALSE;
		retval = 3;
		break;
	  case 'D':
		*nval = gtk_combo_box_get_active(GTK_COMBO_BOX(w));
		retval = 2;
		break;
	  case 'E':
		//widget is a GtkComboBoxEntry with editable text entry field
		txt = gtk_editable_get_chars(GTK_EDITABLE(GTK_BIN(w)->child), 0, -1);
		strcpy(sval, txt);
		retval = 1;
		break;
	  case 'N':
		txt = gtk_editable_get_chars(GTK_EDITABLE(GTK_BIN(w)->child), 0, -1);
		len = strlen(txt);
		for(i = 0; i < len; i++) {
			switch(txt[i]) {
			  case '0': case '1': case '2': case '3': case '4': case '5': 
			  case '6': case '7': case '8': case '9':  break;
			  default:  len = 0;  break;
			}
		}
		if(len && txt[len-1] != '*')  *nval = atoi(txt);
		else  *nval = -1;
		retval = 2;
		break;
	  case 'S':
		txt = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
		strcpy(sval, (txt && strlen(txt)) ? txt : "");
		//strcpy(sval, gtk_entry_get_text(GTK_ENTRY(w)));
		retval = 1;
		break;
	  case 'T':
		//txt = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
		textbuffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(w));
		gtk_text_buffer_get_start_iter(textbuffer, &start);
		gtk_text_buffer_get_end_iter(textbuffer, &end);
		txt = gtk_text_buffer_get_text(textbuffer, &start, &end, FALSE);
		strcpy(sval, (txt && strlen(txt)) ? txt : "");
		for(i = 0; i < strlen(sval); i++)  if(sval[i] == 10 || sval[i] == 13)  sval[i] = ' ';
		retval = 1;
		break;
	  default:
		g_print("Error: AppGetOptValue() - Option Type Not Supported: %s\n", Options[ndx][2]); 
	 	break;
	}
	if(txt)  g_free(txt);
	return(retval);
}

void AppReadWriteConfig (int writeflag, int project)
{
	GtkWidget *w;
	char *sval, line[8193] = "", configfile[260], configtmp[260], name[80];
	char *tstrs[2] = { "opdlg_text_tmp_urls", "opdlg_text_filters_file" };
	char *estrs[2] = { "opdlg_entry_url_listfile", "opdlg_entry_filters_file" };
	int i, nval = 0, pos = 0, scnt;
	gboolean bval = FALSE;
	FILE *infile, *outfile;
	sprintf(configfile, "%s/.ghttrack.cfg", getenv("HOME"));
	sprintf(configtmp, "%s/.ghttrack.tmp", getenv("HOME"));
	sval = g_malloc0(8192);
	if(writeflag) {
		if((rename(configfile, configtmp)) < 0) {
			AppSetStatusText(0, "Error: AppReadWriteConfig(), rename(%s, %s) failed", g_basename(configfile), g_basename(configtmp));
			g_print("Error: AppReadWriteConfig(), rename(%s, %s) failed\n", configfile, configtmp);
			g_free(sval);
			return;
		}
		if((infile = fopen(configtmp, "rb")) == NULL) {
			AppSetStatusText("AppReadWriteConfig(%d), fopen(%s, \"rb\") failed, using defaults, infile=%p", writeflag, configfile, infile);
			g_print("AppReadWriteConfig(%d), fopen(%s, \"rb\") failed, using defaults, infile=%p\n", writeflag, configfile, infile);
			g_free(sval);
			return;
		}
		if((outfile = fopen(configfile, "w+")) == NULL) {
			AppSetStatusText("Error: AppReadWriteConfig(%d), fopen(%s, \"w+\") failed, fn=%p", writeflag, configfile, outfile);
			g_print("Error: AppReadWriteConfig(%d), fopen(%s, \"w+\") failed, fn=%p\n", writeflag, configfile, outfile);
			g_free(sval);
			fclose(infile);
			return;
		}
		ReadLine(infile, line, 80);
		fprintf(outfile, "%s\n", CFGID01);
		ReadLine(infile, line, 80);
		fprintf(outfile, "CurProject=%d\n", Gbl->CurProject);
		for(scnt = 0; scnt < 10; scnt++) {
			if(scnt == Gbl->CurProject) {
				for(i = 0; i < OPTIONCOUNT; i++) {
					ReadLine(infile, line, 8192);
					AppGetOptValue(Options[i][0], sval, &nval, &bval);
					switch(Options[i][2][0]) {
					  case 'E': case 'S': case 'T':  fprintf(outfile, "%d%s=%s\n", scnt, Options[i][0], sval);  break;
					  case 'N': case 'D':  fprintf(outfile, "%d%s=%d\n", scnt, Options[i][0], nval);  break;
					  case 'B':  fprintf(outfile, "%d%s=%d\n", scnt, Options[i][0], bval);  break;
					  default: break;
					}
				}
			}
			else {
				for(i = 0; i < OPTIONCOUNT; i++) {
					ReadLine(infile, line, 8192);
					fprintf(outfile, "%s\n", line);
				}
			}
		}
		fclose(outfile);
		fclose(infile);
		remove(configtmp);
	}
	else {
		if((infile = fopen(configfile, "rb")) == NULL) {
			AppSetStatusText("AppReadWriteConfig(%d), fopen(%s, \"rb\") failed, using defaults, infile=%p", writeflag, configfile, infile);
			g_print("AppReadWriteConfig(%d), fopen(%s, \"rb\") failed, using defaults, infile=%p\n", writeflag, configfile, infile);
			g_free(sval);
			return;
		}
		for(i = 0; i < OPTIONCOUNT; i++) {
			sprintf(name, "%d%s", project, Options[i][0]);
			switch(Options[i][2][0]) {
			  case 'E': case 'S': case 'T': GetConfigString(infile, name, sval, ""); break;
			  case 'N': case 'D': GetConfigInt(infile, name, &nval, 0); break;
			  case 'B': GetConfigBool(infile, name, &bval, FALSE); break;
			  default: break;
			}
			w = lookup_widget(GTK_WIDGET(App1), Options[i][0]);
			switch(Options[i][2][0]) {
			  case 'B':
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), bval);
				break;
			  case 'D':
				gtk_combo_box_set_active(GTK_COMBO_BOX(w), nval);
				break;
			  case 'E':
				gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), sval);
				break;
			  case 'N':
				if(nval == -1)  strcpy(sval, "");
				else  sprintf(sval, "%d", nval);
				gtk_entry_set_text(GTK_ENTRY(GTK_BIN(w)->child), sval);
				break;
			  case 'S':
				gtk_entry_set_text(GTK_ENTRY(w), sval);
				break;
			  case 'T':
				for(pos = 0; pos < strlen(sval); pos++)  if(sval[pos] == ' ')  sval[pos] = 10;
				TextViewSetText(w, sval);
			  default: break;
			}
		}
		sprintf(name, "%d%s", project, "opdlg_rb_use_listfile");
		GetConfigBool(infile, name, &bval, FALSE);
		fclose(infile);
		w = lookup_widget(GTK_WIDGET(App1), bval ? "opdlg_rb_use_listfile" : "opdlg_rb_use_list");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
		for(i = 0; i < 2; i++) {
			w = lookup_widget(GTK_WIDGET(App1), tstrs[i]);
			AppGetOptValue(estrs[i], sval, &nval, &bval);
			if(strlen(sval)) {
				if(!ReadWriteWidgetText(FM_READ, w, sval)) {
					TextViewSetText(w, "");
					w = lookup_widget(GTK_WIDGET(App1), estrs[i]);
					gtk_entry_set_text(GTK_ENTRY(w), "");
				}
			}
			else  TextViewSetText(w, "");
		}
	}
	g_free(sval);
}

void AppRunCommand ()
{
	GtkWidget *w;
	ChildPid = fork();
	if(ChildPid == 0) {
		execl("/bin/bash", "/bin/bash", "-c", Gbl->CmdLine, NULL);
		_exit(EXIT_SUCCESS);
	}
	else if(ChildPid < 0) {
		AppSetStatusText("Error: Fork Failed. Could Not Execute Command. Returning...");
		g_print("Error: Fork Failed. Could Not Execute Command. Returning...\n");
		return;
	}
	AppSetStatusText("Started HTTrack engine (PID=%d).  Re-read log periodically to check progress.", ChildPid);
	w = lookup_widget(GTK_WIDGET(App1), "main_notebook");
	gtk_notebook_set_page(GTK_NOTEBOOK(w), 3);
	//on_logdlg_btn_readlog_clicked(NULL, NULL);
}

void AppRunFileDialog (int mode, char *title)
{
	GtkWidget *w = NULL;
	GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
	Gbl->FileDlgMode = mode;
	if(FileDlg == NULL)  FileDlg = create_FileDlg();
	gtk_window_set_modal(GTK_WINDOW(FileDlg), TRUE);
	gtk_window_set_title(GTK_WINDOW(FileDlg), title);
	switch(mode) {
	  case FD_GET_DEST_LOGS:
	  case FD_GET_DEST_PATH:
		action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
		break;
	  case FD_FILTERS_READ:
	  case FD_URL_LIST_READ:
		action = GTK_FILE_CHOOSER_ACTION_OPEN;
		break;
	  case FD_FILTERS_WRITE:
	  case FD_URL_LIST_WRITE:
		action = GTK_FILE_CHOOSER_ACTION_SAVE;
		break;
	  default:
		g_print("AppRunFileDialog(%d, %s), Unknown Mode\n", mode, title);
		break;
	}
	w = lookup_widget(GTK_WIDGET(FileDlg), "filedlg_filechooser");
	gtk_file_chooser_set_action(GTK_FILE_CHOOSER(w), action);
	gtk_widget_show(FileDlg);
}

void AppSetStatusText (char *txt, ...)
{
	GtkWidget *w;
	char msg[1024];
	va_list args;
	va_start(args, txt);
	vsprintf(msg, txt, args);
	va_end(args);
	w = lookup_widget(GTK_WIDGET(App1), "app1_statusbar");
	gtk_label_set_text(GTK_LABEL(w), msg);
}

void AppShowSettings ()
{
	GtkWidget *w;
	GtkTextBuffer *buf;
	int i, n, ndx, len, nval = 0, prjnum, cmderror = 0;
	int hval = 0, ndx_mime, ndx_rules, indexn = 0, indexh = 0, rb[2];
	char cmdsw[24], sval[4096], tmp[80], temp[4096], line[1024], *stmp;
	gboolean bval = FALSE, use = FALSE, got_max_fsize = FALSE;
	w = lookup_widget(GTK_WIDGET(App1), "app1_text_project_info");
	TextViewSetText(w, "");
	buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(w));
	strcpy(Gbl->CmdLine, "httrack ");
	// --- Project -----------------------------------------------------------
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_combo_project");
	prjnum = gtk_combo_box_get_active(GTK_COMBO_BOX(w));
	if(!prjnum)  sprintf(temp, "\nProject:  %s\n\n", DEFAULTPROJECT);
	else  sprintf(temp, "\nProject:  %s %d\n\n", CUSTOMPROJECT, prjnum);
 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
	AppGetOptValue("opdlg_cb_run_quiet", sval, &nval, &bval);
	if(bval) {
		AppGetOptSwitch("opdlg_cb_run_quiet", cmdsw);
		sprintf(temp, "%s ", cmdsw);
		strcat(Gbl->CmdLine, temp);
		strcpy(line, "Run HTTrack In Quiet Mode\n\n");
	 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
	}
	// --- URL List ----------------------------------------------------------
	AppGetOptValue("opdlg_rb_use_listfile", sval, &nval, &use);
	if(use) {
		AppGetOptValue("opdlg_entry_url_listfile", sval, &nval, &bval);
		sprintf(temp, "%s%s ", AppGetOptSwitch("opdlg_entry_url_listfile", cmdsw), sval);
		strcat(Gbl->CmdLine, temp);
		sprintf(temp, "Use URL List File:  %s\n", sval);
	 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
	}
	else {
		AppGetOptValue("opdlg_text_urls", sval, &nval, &bval);
		if(strlen(sval) < 6)  cmderror++;
		for(i = 0; i < strlen(sval); i++)  if(sval[i] == 10 || sval[i] == 13)  sval[i] = ' ';
		sprintf(temp, "%s ", sval);
		strcat(Gbl->CmdLine, temp);
	}
	strcpy(temp, "List Of URLs To Fetch:\n----------------------------------------\n");
 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
	if(use) {
		// Get URLs from file instead of textbox
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_entry_url_listfile");
		strcpy(temp, gtk_entry_get_text(GTK_ENTRY(w)));
		w = lookup_widget(GTK_WIDGET(App1), "opdlg_tmptext");
		ReadWriteWidgetText(FM_READ, w, temp);
		stmp = TextViewGetText(w);
	 	gtk_text_buffer_insert_at_cursor(buf, stmp, strlen(stmp));
		g_free(stmp);
	}
	else {
		// Get URLs from textbox
		AppGetOptValue("opdlg_text_urls", sval, &nval, &bval);
		for(i = 0; i < strlen(sval); i++)  if(sval[i] == ' ') sval[i] = '\n';
	 	gtk_text_buffer_insert_at_cursor(buf, sval, strlen(sval));
	}
	strcpy(temp, "\n----------------------------------------\n");
 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
	// --- Destination Paths -------------------------------------------------
	AppGetOptValue("opdlg_entry_destination_path", sval, &nval, &bval);
	if(!strlen(sval))  cmderror++;
	sprintf(line, "Destination Path:  %s\n", sval);
 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
	sprintf(temp, "%s%s", AppGetOptSwitch("opdlg_entry_destination_path", cmdsw), sval);
	strcat(Gbl->CmdLine, temp);
	AppGetOptValue("opdlg_cb_destpath_logs", sval, &nval, &use);
	if(use) {
		AppGetOptValue("opdlg_entry_destpath_logs", sval, &nval, &bval);
		sprintf(line, "Path For Logs And Cache:  %s\n\n", sval);
 		gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
		sprintf(temp, ",%s ", sval);
		strcat(Gbl->CmdLine, temp);
	}
	else {
 		gtk_text_buffer_insert_at_cursor(buf, "\n", 1);
		strcat(Gbl->CmdLine, " ");
	}
	// --- Proxy Server ------------------------------------------------------
	AppGetOptValue("opdlg_cb_use_proxy", sval, &nval, &use);
	if(use) {
		AppGetOptValue("opdlg_entry_proxy", sval, &nval, &bval);
		AppGetOptValue("opdlg_entry_proxy_port", tmp, &nval, &bval);
		sprintf(temp, "Use Proxy Server:  %s:%s\n", sval, tmp);
	 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
		AppGetOptSwitch("opdlg_entry_proxy", cmdsw);
		sprintf(temp, "%s%s:%s ", cmdsw, sval, tmp);
		strcat(Gbl->CmdLine, temp);
		AppGetOptValue("opdlg_cb_no_ftp_proxy", sval, &nval, &bval);
		sprintf(temp, "Use Proxy for FTP:  %s\n\n", bval ? "No" : "Yes");
	 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
		if(bval) {
			sprintf(temp, "%s ", AppGetOptSwitch("opdlg_cb_no_ftp_proxy", cmdsw));
			strcat(Gbl->CmdLine, temp);
		}
	}
	// --- Run As User and/or Hostname ---------------------------------------
	AppGetOptValue("opdlg_cb_run_as_user", sval, &nval, &bval);
	if(bval) {
		AppGetOptValue("opdlg_entry_run_as_user", sval, &nval, &bval);
		sprintf(temp, "Run As User:  %s\n\n", sval);
	 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
		sprintf(temp, "%s%s ", AppGetOptSwitch("opdlg_entry_run_as_user", cmdsw), sval);
		strcat(Gbl->CmdLine, temp);
	}
	AppGetOptValue("opdlg_cb_use_hostname", sval, &nval, &bval);
	if(bval) {
		AppGetOptValue("opdlg_entry_hostname", sval, &nval, &bval);
		sprintf(temp, "Use Hostname:  %s\n\n", sval);
	 	gtk_text_buffer_insert_at_cursor(buf, temp, strlen(temp));
		sprintf(temp, "%s%s ", AppGetOptSwitch("opdlg_entry_hostname", cmdsw), sval);
		strcat(Gbl->CmdLine, temp);
	}
	// --- MIME Associations -------------------------------------------------
	ndx = AppGetOptIndex("opdlg_cb_assume_mime");
	AppGetOptValue(Options[ndx][0], sval, &nval, &bval);
	if(bval) {
		sprintf(line, "%s\n", Options[ndx][3]);
	 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
		strcpy(line, "MIME Associations:\n");
	 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
		strcat(Gbl->CmdLine, AppGetOptSwitch(Options[ndx][0], cmdsw));
		temp[0] = 0;
		ndx = AppGetOptIndex("opdlg_entry_mime1a");
		for(i = ndx; i < ndx+16; i += 2) {
			AppGetOptValue(Options[i][0], sval, &nval, &bval);
			AppGetOptValue(Options[i+1][0], tmp, &nval, &bval);
			if(strlen(sval)) {
				sprintf(line, "  %s=%s\n", sval, tmp);
			 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
				if(i == ndx)  sprintf(temp, "\"%s=%s", sval, tmp);
				else  sprintf(temp, ";%s=%s", sval, tmp);
				strcat(Gbl->CmdLine, temp);
			}
		}
	 	gtk_text_buffer_insert_at_cursor(buf, "\n", 1);
		strcat(Gbl->CmdLine, "\" ");
	}
	// --- Scan Rules --------------------------------------------------------
	AppGetOptValue("opdlg_cb_use_filters_list", sval, &nval, &bval);
	if(bval) {
		AppGetOptValue("opdlg_text_filters_list", sval, &nval, &bval);
		for(i = 0; i < strlen(sval); i++) {
			if(sval[i] == 10 || sval[i] == 13)  sval[i] = ' ';
		}
		sprintf(line, "Use Scan Rules List: %s\n", sval);
	 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
		sprintf(temp, "%s ", sval);
		strcat(Gbl->CmdLine, temp);
	}
	else {
		AppGetOptValue("opdlg_cb_use_filters_file", sval, &nval, &bval);
		if(bval) {
			AppGetOptValue("opdlg_entry_filters_file", sval, &nval, &bval);
			sprintf(line, "Use Scan Rules File: %s\n", sval);
	 		gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
			sprintf(temp, "%s%s ", AppGetOptSwitch("opdlg_entry_filters_file", cmdsw), sval);
			strcat(Gbl->CmdLine, temp);
		}
	}
	// --- All Other Options -------------------------------------------------
	indexn = AppGetOptIndex("opdlg_entry_max_size_nonhtml");
	indexh = AppGetOptIndex("opdlg_entry_max_size_html");
	ndx_mime = AppGetOptIndex("opdlg_cb_assume_mime");
	ndx_rules = AppGetOptIndex("opdlg_cb_use_filters_list");
	rb[0] = AppGetOptIndex("opdlg_rb_preset_local_structure");
	rb[1] = AppGetOptIndex("opdlg_rb_custom_local_structure");
	for(i = AppGetOptIndex("opdlg_entry_mirror_depth"); i < OPTIONCOUNT; i++) {
		if((i >= ndx_mime) && (i < (ndx_mime + 17)))  continue;	// MIME already handled
		if((i >= ndx_rules) && (i < (ndx_rules + 4)))  continue;	// Scan Rules already handled
		if(WidgetIsEnabled(App1, Options[i][0]) != 1)  continue;	// Option not active
		AppGetOptValue(Options[i][0], sval, &nval, &bval);
		AppGetOptSwitch(Options[i][0], cmdsw);
		switch(Options[i][2][0]) {
		  case 'B':
			if(bval == FALSE)  continue;
			if(strlen(cmdsw))  sprintf(temp, "%s ", cmdsw);
			sprintf(line, "%s\n", Options[i][3]);
			break;
		  case 'D':
			if(nval == 0)  continue;
			if(!strcmp(Options[i][0], "opdlg_combo_remove_host"))  sprintf(temp, "%s%d ", cmdsw, RemHost[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_preset_lstruct"))  sprintf(temp, "%s%d ", cmdsw, Preset[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_check_doc_type"))  sprintf(temp, "%s%d ", cmdsw, CheckDoc[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_follow_robots_txt"))  sprintf(temp, "%s%d ", cmdsw, FollowRobots[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_scan_mode"))  sprintf(temp, "%s%d ", cmdsw, ScanMode[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_create_logs"))  sprintf(temp, "-%c ", LogType[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_travel_mode"))  sprintf(temp, "-%c ", TravelMode[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_global_travel_mode"))  sprintf(temp, "-%c ", TravelModeG[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_cache_updates"))  sprintf(temp, "%s%d ", cmdsw, CacheUd[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_long_names"))  sprintf(temp, "%s%d ", cmdsw, LongNames[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_keep_original"))  sprintf(temp, "%s ", KeepOriginal[nval]);
			else if(!strcmp(Options[i][0], "opdlg_combo_internet_protocol"))  sprintf(temp, "%s%d ", cmdsw, InternetProtocol[nval]);
			else { g_print("AppGetCommandLine(), Unknown Combo: %s\n", Options[i][0]); continue; };
			sprintf(tmp, "opdlg_combo_%s", Options[i][0]+12);
			if((w = lookup_widget(GTK_WIDGET(App1), tmp)) == 0)  g_print("Widget Not Found:\n\n%s\n", tmp);
			else  sprintf(line, "%s:  %s\n", Options[i][3], gtk_entry_get_text(GTK_ENTRY(GTK_BIN(w)->child)));
			break;
		  case 'N':
			if(nval == -1)  continue;
			if(i == indexn || i == indexh) {
				if(got_max_fsize == FALSE) {
					AppGetOptValue("opdlg_entry_max_size_nonhtml", sval, &nval, &bval);
					AppGetOptValue("opdlg_entry_max_size_html", sval, &hval, &bval);
					AppGetOptSwitch("opdlg_entry_max_size_nonhtml", cmdsw);
					if((hval != -1) && (nval == -1)) {
						nval = 100000;
						sprintf(temp, "%s%d,%d ", cmdsw, nval, hval);
						sprintf(line, "%s:  %d\n%s:  %d\n", Options[indexh][3], hval, Options[indexn][3], nval);
					}
					else if((nval != -1) && (hval == -1)) {
						sprintf(temp, "%s%d ", cmdsw, nval);
						sprintf(line, "%s:  %d\n", Options[indexn][3], nval);
					}
					else {
						sprintf(temp, "%s%d,%d ", cmdsw, nval, hval);
						sprintf(line, "%s:  %d\n%s:  %d\n", Options[indexh][3], hval, Options[indexn][3], nval);
					}
					got_max_fsize = TRUE;
				}
				else  continue;
			}
			else {
				sprintf(temp, "%s%d ", cmdsw, nval);
				sprintf(line, "%s:  %d\n", Options[i][3], nval);
			}
			break;
		  case 'E': case 'S': case 'T':
			len = strlen(sval);
			if(!len)  continue;
			for(n = 0, bval = FALSE; n < len; n++) {
				switch(sval[n]) {
				  case ' ':  case '\%':  case '@':  case '#': case '!':  
					bval = TRUE;  break;
				  default:  break;
				}
			}
			if(bval)  sprintf(temp, "%s\"%s\" ", cmdsw, sval);
			else  sprintf(temp, "%s%s ", cmdsw, sval);
			sprintf(line, "%s:  %s\n", Options[i][3], sval);
			break;
		  default:
			g_print("AppGetCommandLine(), Unknown Type: '%c' in element %d\n", Options[i][2][0], i);
			continue;
			break;
		}
		if(i != rb[0] && i != rb[1])  strcat(Gbl->CmdLine, temp);
	 	gtk_text_buffer_insert_at_cursor(buf, line, strlen(line));
	}
	if(cmderror)  strcpy(Gbl->CmdLine, "");
}

int AppValidList (int list, char *fname)
{
	GtkWidget *w;
	gchar *txt;
	int is_ok = 0;
	w = lookup_widget(GTK_WIDGET(App1), "opdlg_tmptext");
	ReadWriteWidgetText(FM_READ, w, fname);
	txt = TextViewGetText(w);
	if(strlen(txt)) {
		if((list == LIST_FILTERS) && ((txt[0] == '+') || (txt[0] == '-')))  is_ok++;
		else if(list == LIST_URLS) {
			if(!strncasecmp(txt, "http://", 7))  is_ok++;
			else if(!strncasecmp(txt, "ftp://", 6))  is_ok++;
		}
	}
	g_free(txt);
	return(is_ok);
}

///// main //////////////////////////////////////////////////////////////////

int main (int argc, char *argv[])
{
	gtk_set_locale();
	gtk_init(&argc, &argv);
	add_pixmap_directory(PACKAGE_DATA_DIR "/" PACKAGE "/pixmaps");
	Gbl = g_malloc0(sizeof(TGlobalStruct));
	FileDlg = create_FileDlg();
	App1 = create_App1();
	gtk_widget_show(App1);
	gtk_main();
	g_free(Gbl);
	return(0);
}

